import enum
from asyncio import BaseEventLoop, AbstractEventLoop
from typing import Callable, Union, Optional, Type

from NonameOrm.Model.DataModel import DataModel


class GeneratorRunner:
    """
          数据任务执行器

          :return:
          """

    loop: AbstractEventLoop

    def __init__(self, loop: AbstractEventLoop): ...

    def run_group(self, *args): ...

    def run(self): ...

    async def arun(self):
        """
               执行异步创建

               :return:
       """
        ...


class RunnerTask:
    """
        数据生成任务
    """

    _model: Type[DataModel]
    generator: Callable
    count: int

    def __init__(
            self,
            model: Union[Type[DataModel], None],
            generator: Callable
    ): ...

    @property
    def model(self) -> Type[DataModel]: ...


def create_task(
        generator: Callable,
        model: Optional[Type[DataModel]] = None,
        count: Optional[int] = 1,
        group: Union[str, enum] = None
): ...


def data_task(
        model: Optional[Type[DataModel]] = None,
        count: Optional[int] = 1,
        group: Union[str, enum] = None
):...





# from typing import Coroutine, Callable
#
from NonameOrm.Error.DBError import DBInstanceCreateError, DBInstanceError

cdef public DB instance = None

cdef class DB:
    def __init__(self, *args, **kwargs):
        if not kwargs.get('cr'):
            raise DBInstanceCreateError()
        self._connector = kwargs.get('connector')
        global loop

    @classmethod
    def create(cls, *args, **kwargs):
        if isinstance(cls.getInstance(), DB):
            raise DBInstanceError()
        kwargs['cr'] = True
        global instance
        instance = DB(cls, *args, **kwargs)
        return instance

    def GenerateTable(self):
        self._connector.GenerateTable()
        return self

    @property
    def connector(self):
        return self._connector

    @property
    def execute(self):
        if self._connector.isAsync:
            return self.connector.asyncProcess
        else:
            return self.connector.process

    @property
    def executeSql(self):
        return self.connector.execute

    @property
    def instance(self):
        return instance

    @property
    def ConnectorType(self):
        return self.connector.Type

    @classmethod
    def getInstance(cls):
        return instance

    @classmethod
    def getInstanceWithCheck(cls):
        if not instance:
            raise
        return instance



import asyncio
import multiprocessing
import threading
from asyncio import AbstractEventLoop

from NonameOrm.DB.Generator cimport  sqlType, BaseSqlGenerator
from NonameOrm.DB.utils import SqliteContextManager
from NonameOrm.Error.DBError import WriteOperateNotInAffairs
from NonameOrm.Ext import generate_table
import sqlite3

from NonameOrm.Model.ModelExcutor cimport AsyncModelExecutor
import logging

from aiomysql import Cursor, DictCursor

_logger = logging.getLogger(__package__)

lock = multiprocessing.Lock()

def dict_factory(cursor, tuple row):
    d = {}
    cdef int idx

    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d

cdef class BaseConnector:
    def process(self, *args, **kwargs):
        pass

    async def asyncProcess(self, *args, **kwargs):
        pass

    def GenerateTable(self):
        pass

    @property
    def paramsHolder(self):
        return '%s'

cdef class Sqlite3Connector(BaseConnector):
    def __init__(self, str path, bint showLog = True):
        self.isAsync = False
        self.init_sqlite(path, showLog)
        self.path = path

    cdef init_sqlite(self, str path, bint showLog):
        self.con = sqlite3.connect(path, timeout=99999, check_same_thread=False)

        if showLog:
            self.con.set_trace_callback(_logger.debug)

    def releaseCon(self, con):
        con.commit()
        pass

    def getCon(self):
        if threading.current_thread() == threading.main_thread():
            return self.con
        else:
            return sqlite3.connect(self.path)

    @property
    def paramsHolder(self):
        return '?'

    cpdef str autoFiledAnnounce(self, col):
        return 'INTEGER '

    def GenerateTable(self):
        generate_table()

    cpdef list getTableNameList(self):
        cdef tuple table

        cur = self.con.execute('select tbl_name from sqlite_master where type = "table";')
        return [table[0] for table in cur.fetchall()]

    def execute(self, str sql, con=None, bint dictCur=False, tuple args=()):
        cur = self.con.cursor()
        sql = sql.replace('%s', '?')

        if not args: args = ()

        cur.execute(sql, args)

        cdef tuple i

        if dictCur:
            # 当需要返回 dict 格式时
            return (dict_factory(cur, i) for i in cur.fetchall())

        return tuple(cur.fetchall())

    def process(self, ModelExecutor executor, con=None):

        # 获取SQL生成器
        cdef BaseSqlGenerator sql = executor.sql

        # 获取数据库链接
        cdef:
            object res
            str sqlTemp
            object data
        sqlTemp, data = sql.Build()

        sqlTemp = sqlTemp.replace('%s', '?')

        cur = self.con.cursor()

        with lock:
            if isinstance(data, tuple):
                cur.execute(sqlTemp, data)
            else:
                cur.executemany(sqlTemp, data)
            self.con.commit()

        if sql.currentType != sqlType.INSERT:
            if sql.currentType == sqlType.SELECT:
                res = [dict_factory(cur, i) for i in cur.fetchall()]
            else:
                res = cur.fetchall()

        else:
            res = cur.lastrowid

        # 处理查询结果
        return executor.process(res)

cdef class AioSqliteConnector(BaseConnector):
    def __init__(self, loop: AbstractEventLoop, **kwargs):
        self.loop = loop
        self.isAsync = True
        loop.run_until_complete(self.init_sqlite(**kwargs))
        self.isUsing = False

    async def init_sqlite(self, path: str, showLog: bool = True):
        import aiosqlite

        self.con = await aiosqlite.connect(path)

        if showLog:
            await self.con.set_trace_callback(_logger.debug)

        # self.con.row_factory = dict_factory

    async def releaseCon(self, con):
        pass

    @property
    def getCon(self):
        return SqliteContextManager(self.con).acquire

    @property
    def paramsHolder(self):
        return '?'

    def autoFiledAnnounce(self, col):
        return 'INTEGER '

    def GenerateTable(self):
        self.loop.run_until_complete(generate_table())

    async def getTableNameList(self):
        cur = await self.con.execute('select tbl_name from sqlite_master where type = "table";')
        return [table['name'] for table in await cur.fetchall()]

    async def execute(self, str sql, con=None, bint dictCur=False, tuple args=()):
        cur = await self.con.cursor()
        cur.useDict = dictCur
        sql = sql.replace('%s', '?')
        await cur.execute(sql, args)

        cdef tuple i

        if dictCur:
            # 当需要返回 dict 格式时
            return [dict_factory(cur, i) for i in await cur.fetchall()]

        return tuple(await cur.fetchall())

    async def asyncProcess(self, *args, **kwargs):
        cdef AsyncModelExecutor executor = kwargs.get('executor')
        assert isinstance(executor, AsyncModelExecutor), '需要传入executor参数,且必须为AsyncModelExecutor实例'

        # 获取SQL生成器
        cdef BaseSqlGenerator sql = executor.sql

        # 获取数据库链接
        cdef:
            object res
            str sqlTemp
            object data
            bint useSelectCon = False

        sqlTemp, data = sql.Build()

        sqlTemp = sqlTemp.replace('%s', '?')

        cur = await self.con.cursor()

        if isinstance(data, tuple):
            await cur.execute(sqlTemp, data)
        else:
            await cur.executemany(sqlTemp, data)

        if sql.currentType != sqlType.INSERT:
            if sql.currentType == sqlType.SELECT:
                res = [dict_factory(cur, i) for i in await cur.fetchall()]
            else:
                res = await cur.fetchall()
        else:
            res = cur.lastrowid

        # 处理查询结果
        return executor.process(res)

cdef class AioMysqlConnector(BaseConnector):
    def __init__(self, loop: AbstractEventLoop, *args, **kwargs) -> None:
        self._config = kwargs
        self.isAsync = True
        self.Type = AioMysql
        self.selectCon = None
        self.loop = loop
        self._pool = loop.run_until_complete(self._init_mysql(*args, **kwargs))

    async def _init_mysql(self, *args, **kwargs) -> None:
        import aiomysql
        pool = await aiomysql.create_pool(*args, **kwargs)
        return pool

    def GenerateTable(self):
        self.loop.run_until_complete(generate_table())
        pass

    async def getTableNameList(self):
        dbName = self._config.get('db')
        res = await self.execute(f"SELECT TABLE_NAME FROM information_schema.`TABLES` WHERE TABLE_SCHEMA = '{dbName}';")
        return [i[0] for i in res]

    @property
    def config(self):
        return self._config

    def autoFiledAnnounce(self, col):
        cdef str typeArgs = str(col.typeArgs).replace(",", "") if str(col.typeArgs).endswith(",)") else str(col.typeArgs) if len(col.typeArgs) else ''
        return (col.targetType + typeArgs + " ") + 'AUTO_INCREMENT'

    @property
    def getCon(self):
        """
        获取链接实例

        :return:
        """
        return self._pool.acquire

    async def getSelectCon(self):
        return await self._pool.acquire()

    def releaseCon(self, con):
        """
        释放链接实例

        :param con: 数据库链接实例
        :return:
        """
        return self._pool.release(con)

    async def execute(self, str sql, con=None, bint dictCur=False, tuple args=()):
        cdef bint userSelectCon = False
        if not con:
            userSelectCon = True
            con = await self.getSelectCon()
        async with con.cursor(DictCursor if dictCur else Cursor) as cur:
            await cur.execute(sql, args)
            res = await cur.fetchall()

        if userSelectCon:
            await con.commit()
            self.releaseCon(con)

        return res

    async def asyncProcess(self, *args, **kwargs):
        # 获取异步执行对象
        cdef AsyncModelExecutor executor = kwargs.get('executor')
        assert isinstance(executor, AsyncModelExecutor), '需要传入executor参数,且必须为AsyncModelExecutor实例'

        # 获取SQL生成器
        cdef BaseSqlGenerator sql = executor.sql

        # 判断非查询行为是否处于事务模式下
        if sql.currentType != sqlType.SELECT and 'con' not in kwargs:
            raise WriteOperateNotInAffairs()

        # 获取数据库链接
        cdef:
            object res
            str sqlTemp
            list data
            bint useSelectCon = False

        if 'con' in kwargs:
            con = kwargs.get('con')
        else:
            useSelectCon = True
            con = await self.getSelectCon()

        async with con.cursor(DictCursor if sql.currentType == sqlType.SELECT else Cursor) as cur:
            sqlTemp, data = sql.Build()

            if isinstance(data, tuple):
                await cur.execute(sqlTemp, data)
            else:
                await cur.executemany(sqlTemp, data)
            if sql.currentType != sqlType.INSERT:
                res = await cur.fetchall()
            else:
                res = cur.lastrowid

        if useSelectCon:
            await con.commit()
            self.releaseCon(con)
        # 处理查询结果
        return executor.process(res)



# cython_ext: language_level=3
from NonameOrm.Model.ModelProperty cimport BaseProperty
from NonameOrm.Model.ModelProperty cimport FilterListCell

cdef enum sqlType:
    SELECT, INSERT, DELETE, UPDATE, NONE

cdef enum JoinType:
    JOIN, LEFT_JOIN, RIGHT_JOIN, INNER_JOIN

cdef class JoinCell:
    cdef:
        JoinType Type
        object key

cdef class BaseSqlGenerator:
    cdef  sqlType currentType
    cdef public tuple Build(self)

cdef class CustomColAnnounce(BaseSqlGenerator):

    cdef str announce

    cdef public tuple Build(self)

cdef class SqlGenerator(BaseSqlGenerator):
    cdef  list selectCol
    cdef  FilterListCell whereCol
    cdef  object target
    cdef  str limit
    cdef:
        list updateCol
        list _joinList
        tuple orderList

    cpdef public SqlGenerator update(self, object target)
    cpdef public SqlGenerator From(self, object target)
    cpdef public tuple Build(self)
    cdef tuple build_update(self)
    cdef tuple build_select(self)
    @staticmethod
    cdef tuple build_where(FilterListCell whereCol)
    cdef tuple build_insert(self)
    @staticmethod
    cdef str build_order(tuple orderList)
    cdef tuple build_delete(self)
    @staticmethod
    cdef str _getWhereCellStr(FilterListCell cur, list params)
    cdef str build_join(self)

cdef class TableGenerator(BaseSqlGenerator):
    cdef object model
    cpdef public tuple Build(self)
    @staticmethod
    cdef str build_col(BaseProperty col)


class BaseOrmError(BaseException):
    msg: str

    def __init__(self, *args, **kwargs):
        self._process(*args, **kwargs)
        super().__init__(self.msg, *args, **kwargs)

    def _process(self, *args, **kwargs) -> None:
        pass



# cython_ext: language_level=3
# cython: c_string_type=unicode, c_string_encoding=utf8
cdef class DictPlus(dict):
    __slots__ = ()

    def __getattribute__(self, item):
        try:
            return self[item]
        except Exception:
            return object.__getattribute__(self, item)

    def __getattr__(self, item):
        try:
            return self[item]
        except Exception:
            return object.__getattribute__(self, item)

    def __setattr__(self, key, value):
        self.__setitem__(key,value)





# cython_ext: language_level=3
# cython: c_string_type=unicode, c_string_encoding=utf8
import logging
from typing import List

from NonameOrm.DB.Generator cimport SqlGenerator, sqlType
from NonameOrm.Error.DBError import DeleteWithOutPrimaryKeyError, UpdateWithOutPrimaryKeyError

from NonameOrm.Model.DataModel cimport ModelInstance, InstanceList
from NonameOrm.Model.ModelProperty import ForeignType
from pymysql import IntegrityError

from .ModelProperty cimport *
from NonameOrm.DB.DB cimport DB

cdef class BaseModelExecutor:
    def __init__(self, model, work=None):
        self.db = DB.getInstance()
        self.model = model
        self.executed = False
        if work and hasattr(work, 'con'):
            self.work = work.con
        else:
            self.work = work
        self.sql = SqlGenerator().From(model)
        self.__dict__ = {}

    def reset(self):
        if not self.executed:
            return
        self.executed = False
        self.sql = SqlGenerator().From(self.model)
        return self

    def editSql(self, callback):
        callback(self.sql)
        return self

    cdef FilterListCell instanceToFilter(self, ModelInstance instance):
        cdef:
            int i
            BaseProperty cur
            FilterListCell f_list = FilterListCell('1')
        for i in range(len(self.model.col) - 1):
            cur = (<list> self.model.col)[i]
            if not instance.get(cur.name):
                continue
            f_list.append(FilterListCell(cur.name, col=cur), Relationship.AND) \
                .append(FilterListCell(instance.get(cur.name)), Relationship.EQUAL)
        return f_list

    @property
    def generator(self) -> SqlGenerator:
        return self.sql


    def find(self, *cols: List[BaseProperty]):
        pass

    def save(self, instance: ModelInstance):
        pass

    def update(self, instance: ModelInstance):
        pass

    def execute(self):
        pass

    def findAnyMatch(self, instance, int limit=0, int offset=0 ):

        self.reset()
        self.sql.select(*self.model.col)
        self.sql.where(self.instanceToFilter(instance))
        if limit > 0:
            self.sql.limit(limit, offset)
        return  self.execute()

    def findAllBy(self, FilterListCell Filter=None):
        self.reset()
        self.sql.select(*self.model.col) \
            .where(Filter)
        return  self.execute()

    def find(self, *cols: List[BaseProperty]):
        self.reset()
        self.sql.select(*cols)
        return self

    def By(self, FilterListCell Filter = None):
        self.sql.where(Filter)
        return self.execute()

    cdef InstanceList processSelect(self, list res):
        cdef:
            BaseProperty col
            InstanceList instances = InstanceList()
            int i
        for i in range(len(res)):
            instances.append(self.model(res[i]))
        return instances

    cdef processInsert(self, int res):
        from NonameOrm.Model.DataModel import MiddleDataModel
        if issubclass(self.model, MiddleDataModel):
            return
        if 'instance' in self.__dict__:
            if not self.__dict__['instance'][self.model.pkName] and res:
                self.__dict__['instance'][self.model.pkName] = res
            return self.__dict__['instance']
        return self.model({self.model.pkName: res})

    def delete(self, instance, filter=None):
        if not filter and not instance[self.model.pkName]:
            raise DeleteWithOutPrimaryKeyError()
        self.sql.delete(self.model)
        if filter:
            self.sql.where(filter)
        else:
            self.sql.where(self.model.pkCol == instance[self.model.pkName])
        return self.execute()

    cdef process(self, object res):
        if self.sql.currentType == sqlType.INSERT:
            return self.processInsert(<int> res)
        elif self.sql.currentType == sqlType.SELECT:
            return self.processSelect(<list> res)
        elif self.sql.currentType == sqlType.UPDATE:
            return self.__dict__.get('instance')
        else:
            return None


cdef class ModelExecutor(BaseModelExecutor):

    def findListForeignKey(self, InstanceList instances, bint deep=False, dict path = None, int clevel=1, int level=3, str before=None):
        if path is None:
            path = dict()
        cdef int i
        for i in range(len(instances)):
             self.findForeignKey(instances[i], deep=True, path=path, clevel=clevel, level=level)

    def findForeignKey(self, ModelInstance instance, bint deep=False, dict path = None, int clevel=1, int level=3, str before=None):
        if path is None:
            path = dict()
        cdef:
            InstanceList res
            ModelInstance ins
            int i

        if deep and clevel == 1:
            # 若当前为深查找入口时将当前实例放入路径内
            path[self.model.tableName] = {instance[self.model.pkName]: instance}
        elif deep and not clevel == 1:
            # 若当前非深查找入口，则增加深度
            clevel += 1
            if clevel > level:
                # 检查当前深度
                return

        for fk in self.model.fk:
            # 循环查找
            if before == fk.directTarget.tableName:
                continue
            if fk.Type == ForeignType.MANY_TO_MANY:
                # 若外键类型为多对多
                if fk.directTarget.tableName in path and instance[fk.bindCol.name] in path[fk.directTarget.tableName]:
                    # 查看当前外键关联下是否有已查询过的内容，有则直接返回
                    logging.info(f'当前外键{fk.name}已经查询过')
                    instance[fk.name] = path[fk.directTarget.tableName][instance[fk.bindCol.name]]
                    continue
                else:
                    # 通过join中间表查询目标外键表
                    exc: AsyncModelExecutor = fk.directTarget.getExecutor()
                    exc.sql.join(getattr(fk.directTarget, fk.owner.tableName))
                    instance[fk.name] = exc.findAllBy(
                            getattr(fk.middleModel, fk.owner.tableName + '_id') == instance[fk.bindCol.name])

                    # 将当前结果添加进路径
                    if fk.directTarget.tableName in path:
                        path[fk.directTarget.tableName][instance[fk.bindCol.name]] = instance[fk.name]
                    else:
                        path[fk.directTarget.tableName] = {instance[fk.bindCol.name]: instance[fk.name]}

                if deep and len(instance[fk.name]):
                    # 深查找
                    exc: AsyncModelExecutor = instance[fk.name][0].object.getExecutor()
                    for i in range(len(instance[fk.name])):
                        ins = instance[fk.name][i]
                        exc.findForeignKey(ins, deep=True, path=path, clevel=clevel, level=level, before=self.model.tableName)
            else:
                if fk.target.tableName in path and instance[fk.bindCol.name] in path[fk.target.tableName]:
                    instance[fk.name] = path[fk.target.tableName][instance[fk.bindCol.name]]
                    continue
                else:
                    res =  fk.target.getExecutor().findAllBy(
                            fk.targetBindCol == instance[fk.bindCol.name])
                    if res and fk.Type == ForeignType.ONE_TO_ONE and res.len():
                        # 若为一对一关系则直接取结果集中的第一项作为数据放入实例

                        instance[fk.name] = res[0]
                        if deep:
                             instance[fk.name].object.getExecutor().findForeignKey(instance[fk.name], deep=True, path=path, clevel=clevel, level=level, before=self.model.tableName)

                    elif res and fk.Type == ForeignType.ONE_TO_MANY:
                        # 若为一对多关系，则直将结果集作为数据放入实例
                        exc: AsyncModelExecutor = res[0].object.getExecutor()
                        instance[fk.name] = res
                        if deep:
                            for i in range(len(res)):
                                 exc.findForeignKey(res[i], deep=True, path=path, clevel=clevel, level=level)

        return instance



    def save(self, instance):
        self.reset()
        self.__dict__['instance'] = instance


        cdef:
            BaseProperty cur  # 数据列指针
            list insertData = []

        for cur in self.model.col:
            # 判断此列是否在实例内
            if cur.name not in instance or not instance[cur.name]:
                continue
            # 插入数据列表
            insertData.append(cur.insertCell(instance[cur.name]))

        if not insertData:
            return instance
        try:
            self.sql.insert(self.model).values(*insertData)
            res =  self.execute()
        except IntegrityError:
            return  self.update(instance)
        # 外键插入
        from NonameOrm.Model.ModelProperty import ForeignType
        for fk in self.model.fk:
            if fk['name'] not in instance or not instance[fk['name']]:
                continue
            elif fk['Type'] == ForeignType.ONE_TO_ONE:
                self._saveOTO(res, fk, instance)
                continue
            elif fk.Type == ForeignType.ONE_TO_MANY:
                self._saveOTM(fk, instance)
                continue
            elif fk.Type == ForeignType.MANY_TO_MANY:
                self._saveMTM(fk, instance)
                continue
            res[fk['name']] = None if fk.Type == ForeignType else []
            for i in instance[fk['name']]:
                res[fk['name']].append(fk['target'].getExecutor(self.work).save(instance[fk['name']]))
        return res

    def _saveOTO(self, res, fk, instance):
        instance[fk['name']][fk['targetBindCol'].name] = res[fk['bindCol'].name]
        res[fk['name']] =  fk.target.getExecutor(self.work).save(instance[fk['name']])

    def _saveOTM(self, fk, instance):
        exc = fk.target.getExecutor(self.work)
        for model in instance[fk.bindCol.name]:
            exc.reset()
            exc.save(model)

    def _saveMTM(self, fk, instance):
        targetExc = fk.directTarget.getExecutor(self.work)
        middleExc = fk.middleModel.getExecutor(self.work)
        targetFk = fk.middleModel.getOtherFkBy(fk.owner)
        middleExc.delete({fk.targetBindCol.name: instance[fk.bindCol.name]},
                               fk.targetBindCol == instance[fk.bindCol.name])
        for model in instance[fk.name]:
            # await targetExc.save(model)
            middleExc.save({
                    fk.targetBindCol.name: instance[fk.bindCol.name],
                    targetFk.bindCol.name: model[targetFk.targetBindCol.name]
            })

    def update(self, instance: ModelInstance):
        if not instance[self.model.pkName]:
            raise UpdateWithOutPrimaryKeyError()

        self.__dict__['instance'] = instance

        cdef:
            BaseProperty cur
            list params = [cur.updateCell(instance[cur.name]) for cur in self.model.col]

        self.sql.update(self.model).set(*params).where(self.model.pkCol == instance[self.model.pkName])
        from NonameOrm.Model.ModelProperty import ForeignType
        for fk in self.model.fk:
            if fk.name in instance:
                if fk.Type == ForeignType.ONE_TO_ONE:
                    fk.owner.getExecutor(self.work).update(instance[fk.name])
                    continue
                for i in instance[fk.name]:
                    fk.owner.getExecutor(self.work).update(i)

        return self.execute()

    def execute(self):
        self.executed = True
        if self.work:
            return self.db.execute(executor=self, con=self.work)
        else:
            return self.db.execute(executor=self)



cdef class AsyncModelExecutor(BaseModelExecutor):

    async def findListForeignKey(self, InstanceList instances, bint deep=False, dict path = None, int clevel=1, int level=3, str before=None):
        if path is None:
            path = dict()
        cdef int i
        for i in range(len(instances)):
            await self.findForeignKey(instances[i], deep=True, path=path, clevel=clevel, level=level)

    async def findForeignKey(self, ModelInstance instance, bint deep=False, dict path = None, int clevel=1, int level=3, str before=None):
        if path is None:
            path = dict()
        cdef:
            InstanceList res
            ModelInstance ins
            int i

        if deep and clevel == 1:
            # 若当前为深查找入口时将当前实例放入路径内
            path[self.model.tableName] = {instance[self.model.pkName]: instance}
        elif deep and not clevel == 1:
            # 若当前非深查找入口，则增加深度
            clevel += 1
            if clevel > level:
                # 检查当前深度
                return

        for fk in self.model.fk:
            # 循环查找
            if before == fk.directTarget.tableName:
                continue
            if fk.Type == ForeignType.MANY_TO_MANY:
                # 若外键类型为多对多
                if fk.directTarget.tableName in path and instance[fk.bindCol.name] in path[fk.directTarget.tableName]:
                    # 查看当前外键关联下是否有已查询过的内容，有则直接返回
                    logging.info(f'当前外键{fk.name}已经查询过')
                    instance[fk.name] = path[fk.directTarget.tableName][instance[fk.bindCol.name]]
                    continue
                else:
                    # 通过join中间表查询目标外键表
                    exc: AsyncModelExecutor = fk.directTarget.getExecutor()
                    exc.sql.join(getattr(fk.directTarget, fk.owner.tableName))
                    instance[fk.name] = await exc.findAllBy(
                            getattr(fk.middleModel, fk.owner.tableName + '_id') == instance[fk.bindCol.name])

                    # 将当前结果添加进路径
                    if fk.directTarget.tableName in path:
                        path[fk.directTarget.tableName][instance[fk.bindCol.name]] = instance[fk.name]
                    else:
                        path[fk.directTarget.tableName] = {instance[fk.bindCol.name]: instance[fk.name]}

                if deep and len(instance[fk.name]):
                    # 深查找
                    exc: AsyncModelExecutor = instance[fk.name][0].object.getExecutor()
                    for i in range(len(instance[fk.name])):
                        ins = instance[fk.name][i]
                        await exc.findForeignKey(ins, deep=True, path=path, clevel=clevel, level=level, before=self.model.tableName)
            else:
                if fk.target.tableName in path and instance[fk.bindCol.name] in path[fk.target.tableName]:
                    instance[fk.name] = path[fk.target.tableName][instance[fk.bindCol.name]]
                    continue
                else:
                    res = await fk.target.getExecutor().findAllBy(
                            fk.targetBindCol == instance[fk.bindCol.name])
                    if res and fk.Type == ForeignType.ONE_TO_ONE and res.len():
                        # 若为一对一关系则直接取结果集中的第一项作为数据放入实例

                        instance[fk.name] = res[0]
                        if deep:
                            await instance[fk.name].object.getExecutor().findForeignKey(instance[fk.name], deep=True, path=path, clevel=clevel, level=level, before=self.model.tableName)

                    elif res and fk.Type == ForeignType.ONE_TO_MANY:
                        # 若为一对多关系，则直将结果集作为数据放入实例
                        exc: AsyncModelExecutor = res[0].object.getExecutor()
                        instance[fk.name] = res
                        if deep:
                            for i in range(len(res)):
                                await exc.findForeignKey(res[i], deep=True, path=path, clevel=clevel, level=level)

        return instance


    async def save(self, instance):
        self.reset()
        self.__dict__['instance'] = instance


        cdef:
            BaseProperty cur  # 数据列指针
            list insertData = []

        for cur in self.model.col:
            # 判断此列是否在实例内
            if cur.name not in instance or not instance[cur.name]:
                continue
            # 插入数据列表
            insertData.append(cur.insertCell(instance[cur.name]))

        if not insertData:
            return instance
        try:
            self.sql.insert(self.model).values(*insertData)
            res = await self.execute()
        except IntegrityError:
            return await self.update(instance)
        # 外键插入
        from NonameOrm.Model.ModelProperty import ForeignType
        for fk in self.model.fk:
            if fk['name'] not in instance or not instance[fk['name']]:
                continue
            elif fk['Type'] == ForeignType.ONE_TO_ONE:
                await self._saveOTO(res, fk, instance)
                continue
            elif fk.Type == ForeignType.ONE_TO_MANY:
                await self._saveOTM(fk, instance)
                continue
            elif fk.Type == ForeignType.MANY_TO_MANY:
                await self._saveMTM(fk, instance)
                continue
            res[fk['name']] = None if fk.Type == ForeignType else []
            for i in instance[fk['name']]:
                res[fk['name']].append(await fk['target'].getExecutor(self.work).save(instance[fk['name']]))
        return res

    async def _saveOTO(self, res, fk, instance):
        instance[fk['name']][fk['targetBindCol'].name] = res[fk['bindCol'].name]
        res[fk['name']] = await fk.target.getExecutor(self.work).save(instance[fk['name']])

    async def _saveOTM(self, fk, instance):
        exc = fk.target.getExecutor(self.work)
        for model in instance[fk.bindCol.name]:
            exc.reset()
            await exc.save(model)

    async def _saveMTM(self, fk, instance):
        targetExc = fk.directTarget.getExecutor(self.work)
        middleExc = fk.middleModel.getExecutor(self.work)
        targetFk = fk.middleModel.getOtherFkBy(fk.owner)
        await middleExc.delete({fk.targetBindCol.name: instance[fk.bindCol.name]},
                               fk.targetBindCol == instance[fk.bindCol.name])
        for model in instance[fk.name]:
            # await targetExc.save(model)
            await middleExc.save({
                    fk.targetBindCol.name: instance[fk.bindCol.name],
                    targetFk.bindCol.name: model[targetFk.targetBindCol.name]
            })


    async def update(self, instance: ModelInstance):
        if not instance[self.model.pkName]:
            raise UpdateWithOutPrimaryKeyError()

        self.__dict__['instance'] = instance

        cdef:
            BaseProperty cur
            list params = [cur.updateCell(instance[cur.name]) for cur in self.model.col]

        self.sql.update(self.model).set(*params).where(self.model.pkCol == instance[self.model.pkName])
        from NonameOrm.Model.ModelProperty import ForeignType
        for fk in self.model.fk:
            if fk.name in instance:
                if fk.Type == ForeignType.ONE_TO_ONE:
                    await fk.owner.getExecutor(self.work).update(instance[fk.name])
                    continue
                for i in instance[fk.name]:
                    await fk.owner.getExecutor(self.work).update(i)

        return await self.execute()

    async def execute(self):
        self.executed = True
        if self.work:
            return await self.db.execute(executor=self, con=self.work)
        else:
            return await self.db.execute(executor=self)



from .BaseOrmError import *
class DBInstanceError(BaseOrmError):
    msg:str = "DB链接对象实例已存在，不可重新创建"

class DBInstanceCreateError(BaseOrmError):
    msh:str = "请使用DB.create()创建数据库实例"

class WriteOperateNotInAffairs(BaseOrmError):
    msg:str = "写操作需要在事务中执行"

class UpdateWithOutPrimaryKeyError(BaseOrmError):
    msg = "更新数据库实例需要包含主键字段"

class DeleteWithOutPrimaryKeyError(BaseOrmError):
    msg = "删除数据库实例需要包含主键字段"


from NonameOrm.Model.ModelExcutor cimport AsyncModelExecutor,BaseModelExecutor
from NonameOrm.Model.ModelProperty cimport FilterListCell

from NonameOrm.Ext.Dict cimport DictPlus

cdef class Page(DictPlus):
    pass

cdef class PageAble:
    cdef:
        object target
        BaseModelExecutor executor
        FilterListCell filter
        int page, pageSize
        bint deep

    cpdef Page _execute(self)


import enum
from typing import Type, Optional, Union, Callable, Dict, List

from NonameOrm.DB.DB cimport DB
from NonameOrm.Model.DataModel import DataModel
from NonameOrm.Model.DataModel cimport ModelInstance
from NonameOrm.Model.ModelExcutor cimport AsyncModelExecutor, ModelExecutor

cdef list tasks = []  # 任务列表

cdef dict taskGroup = {}  # 任务分组

tasks: List['RunnerTask']
taskGroup: Dict[Union[str, enum], List['RunnerTask']]

cdef class GeneratorRunner:
    cdef object loop

    def __init__(self, loop=None):
        self.loop = loop
        pass

    def run_group(self, *args):
        pass

    def run(self):
        if DB.getInstance().connector.isAsync:
            self.loop.run_until_complete(self.arun())
        else:
            self._run()

    cdef _run(self):
        cdef:
            int index, time
            RunnerTask task
            ModelInstance instance
            ModelExecutor executor

        for index in range(len(tasks)):
            task: RunnerTask = tasks[index]

            executor = task.model.getExecutor()

            for time in range(task.count):
                instance = task.generator(time=time + 1)
                executor.save(instance)

    async def arun(self):
        """
        执行异步创建

        :return:
        """
        cdef:
            int index, time
            RunnerTask task
            ModelInstance instance
            AsyncModelExecutor executor

        for index in range(len(tasks)):
            task: RunnerTask = tasks[index]

            executor = task.model.getExecutor()

            for time in range(task.count):
                instance = task.generator(time=time + 1)
                await executor.save(instance)

cdef class RunnerTask:
    """
        生成任务
    """

    cdef:
        object _model  # 模型对象
        object generator  # 实例生成函数
        int count  # 执行次数

    def __init__(self, model, int count, generator):
        self._model = model
        self.count = count
        self.generator = generator

    @property
    def model(self) -> DataModel:
        if not self._model:
            self._model = self.generator(time=1).object

        return self._model
def data_task(
        model: Optional[Type[DataModel]] = None,
        count: Optional[int] = 1,
        group: Union[str, enum] = None
):
    """
    创建 数据生成任务
    装饰器

    :return:
    """
    def inner(fun):
        create_task(fun, model, count, group)
        return fun
    return inner

def create_task(
        generator: Callable,
        model: Optional[Type[DataModel]] = None,
        count: Optional[int] = 1,
        group: Union[str, enum] = None):
    """
       创建 数据生成任务

       :return:
   """
    cdef RunnerTask task = RunnerTask(model=model, count=count, generator=generator)

    if group:
        if group not in taskGroup: taskGroup[group] = []  # 创建分组
        taskGroup[group].append(task)  # 加入分组

    tasks.append(task)



# cython_ext: language_level=3
from NonameOrm.DB.Generator cimport CustomColAnnounce






cdef class NullDefault:
    pass

cdef class AutoIncrement:
    pass

cdef class BaseProperty:
    cdef public str name
    cdef type _Type
    cdef public bint isPk
    cdef public bint Null
    cdef public object _default
    cdef public object _targetType
    cdef object model
    cdef str define
    cdef public tuple typeArgs

    cpdef public bint sizeChecker(self, object value)
    cpdef public bint verifier(self, object value)
    cdef public str desc(self)
    cdef public str asc(self)

    cpdef FilterListCell setFilter(self, object other, Relationship relationship)

cdef enum Relationship:
    AND
    OR
    EQUAL
    BIGGER
    BIGGER_EQUAL
    SMALLER
    SMALLER_EQUAL
    NOTEQUAL
    NONE
    LIKE

cdef class FilterListCell:
    cdef public object value
    cdef public Relationship relationship
    cdef public FilterListCell next
    cdef public BaseProperty col

    cpdef FilterListCell append(self, FilterListCell cell, Relationship relationship)
    cpdef _add(self, object other, Relationship relationship)




from typing import Type, Optional, List, Tuple, Dict, Union

from NonameOrm.DB.DB import DB
from NonameOrm.Model.ModelExcutor import AsyncModelExecutor, ModelExecutor
from NonameOrm.Model.ModelProperty import BaseProperty, ForeignKey


class ModelInstance(dict):
    def __init__(self, *args, **kwargs): ...

    def __getattribute__(self, name): ...

    def __setattr__(self, key, value): ...


class _DataModel:
    @classmethod
    def instanceBuilder(cls: DataModel, *args, **kwargs) -> ModelInstance: ...


class _DataModelMeta(type):
    def __new__(cls, name, bases: tuple, attrs: dict, **kwargs): ...


class DataModel(_DataModel, ModelInstance, metaclass=_DataModelMeta):
    col: Tuple[BaseProperty]
    mapping: Dict[str, BaseProperty]
    tableName: str
    _db: Type[DB] = DB
    pkName: Optional[str] = None
    pkCol: Optional[BaseProperty] = None
    fk: List[ForeignKey]

    def __new__(cls, *args, **kwargs) -> ModelInstance: ...

    @classmethod
    def getAsyncExecutor(cls, work=None) -> AsyncModelExecutor: ...

    @classmethod
    def getExecutor(cls, work=None) -> Union[AsyncModelExecutor, ModelExecutor]: ...






cdef class _DataModel:
    cdef tuple col
    cdef dict mapping
    cdef modelInstance
    cdef public str tableName

cdef class InstanceList(list):
    cdef int len(self)
    pass

cdef class ModelInstance(dict):
    cdef dict _temp
    pass


from NonameOrm.DB.DB import DB


class SqliteContextManager:
    """
        只为了包装 sqlite 来给上下文管理使用
    """

    def __init__(self, con):
        self.con = con

    def acquire(self):
        if DB.getInstance().connector.isAsync:
            return self._async_acquire()
        return self._acquire()

    async def _async_acquire(self):
        return self.con

    def _acquire(self):
        return self.con

    def __enter__(self):
        return self._acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    async def __aenter__(self):
        return await self._async_acquire()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass



import functools

from NonameOrm.DB.DB import DB

def use_database(fun):
    @functools.wraps(fun)
    async def warp(self, *args, **kwargs):
        async with DB.getInstance().connector.getCon() as self.con:
            try:
                res = await fun(self, *args, **kwargs)
                await self.con.commit()
                return res
            except Exception as e:
                await self.con.rollback()
                raise e

    return warp



import ast
import os
from enum import Enum
from typing import List, Tuple, Dict

import astor
import re

import logging

from case_convert import pascal_case

log = logging.getLogger(__package__)

typeMatcher = re.compile(r'\(.*\)')


async def async_generate_table():
    from NonameOrm.DB.DB import DB
    from NonameOrm.DB.Generator import TableGenerator
    from NonameOrm.Model.DataModel import DataModel, _DataModelMeta, MiddleDataModel

    modelList: List[DataModel] = _DataModelMeta.ModelList
    db = DB.getInstance()
    nameList: List[str] = await db.connector.getTableNameList()

    con = await db.connector.getCon()
    try:
        for model in modelList:
            if model.tableName in nameList or model == DataModel or model == MiddleDataModel:
                continue
            await db.executeSql(*TableGenerator(model).Build(), con)
    except Exception as e:
        await con.rollback()
        raise e
    await con.commit()
    await db.connector.releaseCon(con)


def _generate_table():
    from NonameOrm.DB.DB import DB
    from NonameOrm.DB.Generator import TableGenerator
    from NonameOrm.Model.DataModel import DataModel, _DataModelMeta, MiddleDataModel

    modelList: List[DataModel] = _DataModelMeta.ModelList
    db = DB.getInstance()
    nameList: List[str] = db.connector.getTableNameList()

    con = db.connector.getCon()
    try:
        for model in modelList:
            if model.tableName in nameList or model == DataModel or model == MiddleDataModel:
                continue
            db.executeSql(*TableGenerator(model).Build(), con)
    except Exception as e:
        con.rollback()
        raise e
    con.commit()
    db.connector.releaseCon(con)


def generate_table():
    from NonameOrm.DB.DB import DB

    if DB.getInstance().connector.isAsync:
        return async_generate_table()

    return _generate_table()


class Type(Enum):
    int = 'int'
    bigint = 'bigint'
    float = 'float'
    decimal = 'decimal'
    datetime = 'datetime'
    timestamp = 'timestamp'
    json = 'json'
    varchar = 'varchar'
    text = 'text'
    longtext = 'longtext'
    tinyint = 'tinyint'
    tinytext = 'tinytext'
    bit = 'bit'


class Null(Enum):
    yes = 'yes'
    no = 'no'


class Key(Enum):
    UNI = 'UNI'
    PRI = 'PRI'


class Extra(Enum):
    auto_increment = 'auto_increment'





typeMap: dict = {
    Type.bigint.value: 'IntProperty',
    Type.int.value: 'IntProperty',
    Type.decimal.value: 'FloatProperty',
    Type.json.value: 'JsonProperty',
    Type.float.value: 'FloatProperty',
    Type.text.value: 'StrProperty',
    Type.longtext.value: 'StrProperty',
    Type.tinyint.value: 'IntProperty',
    Type.tinytext.value: 'StrProperty',
    Type.varchar.value: 'StrProperty',
    Type.bit.value: 'BoolProperty',
    Type.timestamp.value: 'TimestampProperty',
    Type.datetime.value: 'TimestampProperty'
}

from NonameOrm.Model.ModelProperty import *

supportTypeMap: Dict[str, type(Enum)] = {
    'IntProperty': strSupportType,
    'FloatProperty': floatSupportType,
    'JsonProperty': jsonSupportType,
    'StrProperty': strSupportType,
    'BoolProperty': boolSupportType,
    'TimestampProperty': timestampSupportType
}


def tableAnnounceToAstModule(colAnnounce: Tuple[dict], tableName: str) -> ast.Module:
    # 创建AST根节点
    module: ast.Module = ast.Module(body=[])
    # 创建导入节点并添加到根节点末尾
    module.body.append(ast.ImportFrom(module='NonameOrm.Model.ModelProperty', names=[ast.alias('*', None)], level=0))
    module.body.append(
        ast.ImportFrom(module='NonameOrm.Model.DataModel', names=[ast.alias('DataModel', None)], level=0))

    # 创建类定义节点
    classDefNode: ast.ClassDef = ast.ClassDef(
        bases=[ast.Name(id='DataModel', ctx=ast.Load())],
        keyword=[],
        name=pascal_case(tableName),
        body=[colAnnounceToAstNode(col) for col in colAnnounce],
        decorator_list=[]
    )
    module.body.append(classDefNode)
    return module


def colAnnounceToAstNode(colAnnounce: dict) -> ast.stmt:
    return ast.Assign(
        targets=[ast.Name(id=colAnnounce['Field'], ctx=ast.Store)],
        value=_getColValueNode(colAnnounce),
        type_comment=''
    )


def _getColValueNode(colAnnounce: dict) -> ast.expr:
    """
    用于创建数据列声明节点

    :param colAnnounce: 数据库列声明字典
    :return: ast.expr 节点
    """
    node: ast.Call = ast.Call(
        func=ast.Name(id=typeMap[typeMatcher.sub('', colAnnounce['Type'])], ctx=ast.Load()),
        args=[],
        keywords=[]
    )
    if colAnnounce['Key'] == Key.PRI.value:
        # 添加主键声明参数
        node.keywords.append(ast.keyword(arg='pk', value=ast.Constant(value=True)))
        if colAnnounce['Extra'] == Extra.auto_increment.value:
            # 判断是否自增字段，是则添加自增声明参数
            node.keywords.append(ast.keyword(arg='default', value=ast.Name(id='auto_increment', ctx=ast.Load())))

    if colAnnounce['Default']:
        # 判断字段是否有默认值，有则添加默认值
        node.keywords.append(ast.keyword(arg='default', value=ast.Constant(value=colAnnounce['Default'])))

    typeArgs = typeMatcher.findall(colAnnounce['Type'])
    if typeArgs:
        # 判断是否有类型参数，有则添加类型参数节点
        node.keywords.append(
            ast.keyword(
                arg='typeArgs',
                value=ast.Tuple(
                    ctx=ast.Load,
                    elts=[ast.Constant(value=x) for i in typeArgs for x in re.findall(r'\d+', i)]
                )
            )
        )
        typeName = typeMatcher.sub('', colAnnounce['Type'])
        # 处理字段类型
        try:
            colType = supportTypeMap[typeMap[typeName]]
            node.keywords.append(
                ast.keyword(
                    arg='targetType',
                    value=ast.Attribute(
                        ctx=ast.Load,
                        attr=typeName,
                        value=ast.Name(
                            ctx=ast.Load,
                            id=colType.__name__
                        )
                    )
                )
            )
        except Exception:
            log.info(f'无匹配类型{typeName},故略过')
            pass

    return node


async def generate_model(filePath: str):
    """
    自动侦测数据库表结构，并生成相应的模型类文件

    :param filePath: 模型类文件生成路径
    :return: void
    """
    from NonameOrm.DB.DB import DB

    db = DB.getInstance()
    dbName = db.connector.config.get('db')
    res: Tuple[Tuple[str]] = await db.executeSql(f"SELECT TABLE_NAME FROM information_schema.`TABLES` WHERE TABLE_SCHEMA = '{dbName}';")
    for i in range(len(res)):
        tableName: str = res[i][0]
        colAnnounce: Tuple[dict] = await db.executeSql(f'desc {tableName}', dictCur=True)
        await _saveAstModuleByTableName(tableName, colAnnounce, filePath)


async def autoGenerate(filePath: str):
    """
    自动匹配本地数据模型与数据库表结构的差异，自动更新

    :param filePath: 模型类文件生成路径
    :return: void
    """
    from NonameOrm.DB.DB import DB

    db = DB.getInstance()
    dbName = db.connector.config.get('db')
    tableTuple: Tuple[Tuple[str]] = await db.executeSql(f"SELECT TABLE_NAME FROM information_schema.`TABLES` WHERE TABLE_SCHEMA = '{dbName}';")
    con = await db.connector.getCon()
    pyFiles = _getAllPyFilesByPath(filePath)
    for (tableName,) in tableTuple:
        colAnnounce: Tuple[dict] = await db.executeSql(f'desc {tableName}', dictCur=True)
        if pascal_case(tableName) + '.py' in pyFiles:
            with open('', 'r') as f:
                await _updateModelAst(ast.parse(f.read()), tableName, colAnnounce)
        else:

            await _saveAstModuleByTableName(tableName, colAnnounce, filePath)


async def _updateModelAst(moduleAst: ast.Module, tableName: str, filePath: str, colAnnounceList: List[dict]):
    """
    用于更新DataModel文件

    :param modelAst:
    :param tableName:
    :param filePath:
    :return:
    """
    colAssignMap: Dict[str:ast.Assign] = _getColAssignFromAstModule(moduleAst)
    for colAnnounce in colAnnounceList:
        if colAnnounce['Field'] in colAssignMap:
            del colAssignMap[colAnnounce['Field']]
            continue

    with open((filePath if filePath.endswith('/') else filePath + '/') + f'{pascal_case(tableName)}.py', '+') as f:
        sourceCode = astor.to_source(moduleAst)
        f.write(sourceCode)


async def _saveAstModuleByTableName(tableName: str, colAnnounce, filePath):
    """
    根据数据库表名自动分析并生成对应的NonameOrm模型文件

    :param tableName: 数据库表名
    :param db: 数据库实例
    :param filePath: 模型文件生成路径
    :return: void
    """
    module: ast.Module = tableAnnounceToAstModule(colAnnounce, tableName)
    with open((filePath if filePath.endswith('/') else filePath + '/') + f'{pascal_case(tableName)}.py', 'w') as f:
        sourceCode = astor.to_source(module)
        f.write(sourceCode)


def _getColAssignFromAstModule(module: ast.Module) -> Dict[str, ast.Assign]:
    """
    从

    :param module:
    :return:
    """
    colAssignMap: Dict[str, ast.Assign] = dict()
    for node in ast.walk(module):
        if isinstance(node, ast.Assign) and node.targets[0].id.endswith('Property'):
            colAssignMap[node.targets[0].id] = node
    return colAssignMap


def _getAllPyFilesByPath(path: str) -> List[str]:
    """
    获取指定路径下所有python文件名

    :param path: 系统路径
    :return: List[str]
    """
    pyFiles = []
    for i, j, k in os.walk(path):
        pyFiles.extend([file for file in k if file.endswith('.py')])
    return pyFiles



from NonameOrm.Model.ModelExcutor cimport ModelExecutor

cdef enum ConnectorType:
    AioMysql


cdef class BaseConnector:
    cdef:
        public object _pool
        ConnectorType Type
        public bint isAsync
        object selectCon
        public dict _config
        int count

cdef class Sqlite3Connector(BaseConnector):

    cdef:
        object con
        str path

    cdef init_sqlite(self, str path, bint showLog)
    cpdef str autoFiledAnnounce(self, col)
    cpdef list getTableNameList(self)

cdef class AioMysqlConnector(BaseConnector):
    cdef :
        object loop
    pass


cdef class AioSqliteConnector(BaseConnector):
    cdef:
        object loop
        object con
        bint isUsing




from typing import Callable, List, Any, Optional

from NonameOrm.DB.DB import DB
from NonameOrm.DB.Generator import SqlGenerator
from NonameOrm.Model.DataModel import InstanceList, ModelInstance, DataModel
from NonameOrm.Model.ModelProperty import BaseProperty, FilterListCell


class BaseModelExecutor:
    sql: SqlGenerator
    model: DataModel
    db: DB
    work: Any

    def __init__(self, model: DataModel, work=None): ...

    def instanceToFilter(self, instance: ModelInstance): ...

    def reset(self) -> BaseModelExecutor: ...

    @property
    def generator(self) -> SqlGenerator: ...

    def findAnyMatch(self, instance: ModelInstance, limit=None, offset=0): ...

    def find(self, *cols: List[BaseProperty]): ...

    def save(self, instance: ModelInstance): ...

    def update(self, instance: ModelInstance): ...

    def editSql(self, callback: Callable[[SqlGenerator], None]) -> BaseModelExecutor: ...

    def execute(self): ...

    def process(self, res): ...


class AsyncModelExecutor(BaseModelExecutor):

    async def findAnyMatch(self, instance: ModelInstance, limit=None, offset=0) -> List: ...

    async def findAllBy(self, Filter: FilterListCell = None): ...

    def find(self, *cols: List[BaseProperty]) -> AsyncModelExecutor: ...

    async def By(self, Fileter: FilterListCell) -> ModelInstance: ...

    async def save(self, instance: ModelInstance): ...

    async def update(self, instance: ModelInstance): ...

    async def execute(self) -> ModelInstance: ...

    async def findForeignKey(self, instance, deep: Optional[bool] = False): ...

    async def findListForeignKey(self, instances: InstanceList, deep: Optional[bool] = False):...



# cython_ext: language_level=3
# cython: c_string_type=unicode, c_string_encoding=utf8
import ast
from typing import List

from NonameOrm.Error.SqlError import *
from NonameOrm.Model.ModelProperty import BaseProperty, FilterListCell, NullDefault
from NonameOrm.Model.ModelProperty cimport Relationship
from NonameOrm.Model.ModelProperty cimport AutoIncrement

cdef dict relationshipMap = {
    Relationship.AND: " AND ",
    Relationship.OR: " OR ",
    Relationship.EQUAL: " = ",
    Relationship.BIGGER: " > ",
    Relationship.SMALLER: " < ",
    Relationship.NOTEQUAL: " != ",
    Relationship.BIGGER_EQUAL: " >= ",
    Relationship.SMALLER_EQUAL: " <= ",
    Relationship.LIKE: " LIKE "
}

cdef dict joinTypeMap = {
    JOIN: "JOIN ",
    LEFT_JOIN: "LEFT JOIN ",
    RIGHT_JOIN: "RIGHT JOIN ",
    INNER_JOIN: "INNER JOIN ",
}

cdef class JoinCell:
    def __init__(self, fk, JoinType joinType = JoinType.JOIN):
        self.Type = joinType
        self.key = fk

cdef class BaseSqlGenerator:
    cdef public tuple Build(self):
        return ()

cdef class CustomColAnnounce(BaseSqlGenerator):
    def __init__(self, str announce):
        self.announce = announce

    cdef public tuple Build(self):
        return self.announce, None

cdef class SqlGenerator(BaseSqlGenerator):
    def __init__(self):
        self.currentType = NONE
        self.limit = ' '
        self.target = None

    def insert(self, model):
        if not self.currentType == NONE:
            raise SqlInStanceError(self.currentType, sqlType.INSERT)
        self.currentType = sqlType.INSERT
        self.From(model)
        return self

    def values(self, *args):
        self.selectCol = list(args)
        return self

    def select(self, *args: List[BaseProperty]) -> SqlGenerator:
        if not self.currentType == NONE:
            raise SqlInStanceError(self.currentType, sqlType.SELECT)
        self.currentType = sqlType.SELECT
        self.selectCol = list(args)
        return self

    def delete(self, target):
        if not self.currentType == NONE:
            raise SqlInStanceError(self.currentType, sqlType.DELETE)
        self.From(target)
        self.currentType = sqlType.DELETE
        return self

    cpdef public SqlGenerator update(self, object target):
        self.currentType = sqlType.UPDATE
        self.From(target)
        return self

    def set(self, *args)-> SqlGenerator:
        if self.currentType == sqlType.NONE:
            raise SetSQLError()
        self.updateCol = list(args)
        return self

    cpdef public SqlGenerator From(self, object target):
        if isinstance(target, str):
            self.target = target
        else:
            from NonameOrm.Model.DataModel import DataModel
            assert isinstance(target, type) and issubclass(target, DataModel), 'Form仅支持字符串或者DataModel子类'
            self.target = target.tableName
        return self

    def where(self, FilterListCell args=None) -> SqlGenerator:
        if self.currentType == sqlType.NONE:
            raise WhereSQLError()
        self.whereCol = args
        return self

    cpdef public tuple Build(self):
        if self.currentType == sqlType.SELECT:
            return self.build_select()
        elif self.currentType == sqlType.UPDATE:
            return self.build_update()
        elif self.currentType == sqlType.INSERT:
            return self.build_insert()
        elif self.currentType == sqlType.DELETE:
            return self.build_delete()

    def Limit(self, int count, int offset) -> SqlGenerator:
        self.limit = " limit %i, %i" % (offset, count)
        return self

    def orderBy(self, *args):
        self.orderList = args
        pass

    def join(self, object foreignKey, joinType: JoinType = JoinType.JOIN) -> SqlGenerator:
        self.joinList.append(JoinCell(foreignKey, joinType))
        return self

    def leftJoin(self, object foreignKey) -> SqlGenerator:
        return self.join(foreignKey, JoinType.LEFT_JOIN)

    def rightJoin(self, object foreignKey) -> SqlGenerator:
        return self.join(foreignKey, JoinType.RIGHT_JOIN)

    def innerJoin(self, object foreignKey) -> SqlGenerator:
        return self.join(foreignKey, JoinType.INNER_JOIN)

    cdef tuple build_delete(self):
        cdef:
            str whereTemp
            list params

        whereTemp, params = SqlGenerator.build_where(self.whereCol)

        return "DELETE FROM " + self.target + " " + whereTemp, tuple(params)

    cdef tuple build_insert(self):
        cdef:
            str insertTemp = "INSERT INTO " + self.target + " "  # INSERT 语句
            dict cur  # 数据指针
            list params = []  # sql参数
            list cols = []
            int i

        # 将插入数据压入列表
        for cur in self.selectCol:
            cols.append((<BaseProperty> cur.get('col')).name)
            params.append(cur.get('value'))

        # 拼接完整sql语句
        insertTemp += "(" + ",".join(cols) + ") values (" + ",".join(['%s' for i in range(len(params))]) + ');'

        return insertTemp, tuple(params)

    cdef tuple build_update(self):
        cdef:
            str updateTemp = "UPDATE " + self.target + " SET "
            dict cur
            list params = []
            str whereTemp
            list whereParams

        for cur in self.updateCol:
            updateTemp += cur['name'] + " = %s,"
            params.append(cur['value'])
        whereTemp, whereParams = SqlGenerator.build_where(self.whereCol)
        return updateTemp[:-1] + whereTemp + ";", tuple(params + whereParams)

    @staticmethod
    cdef str build_order(tuple orderList):
        """
        生成 orderBy 子句
        
        :return: orderBy 子句内容 
        """
        if not orderList or len(orderList):
            return ""

        cdef:
            int i
            int length = len(orderList)
            str orderTemp = "order by "


        for i in range(length):
            orderTemp += orderList[i]
            if not i == length - 1:
                orderTemp += ", "

        return orderTemp

    cdef tuple build_select(self):
        """
        生成select子句
        
        :return: select 子句 
        """
        cdef:
            BaseProperty Property
            str whereTemp
            str selectTemp = "SELECT "
            list params

        selectTemp += ",".join(
            [f"{Property.model.tableName}.{Property.name}" for Property in self.selectCol if Property])
        whereTemp, params = SqlGenerator.build_where(self.whereCol)

        if self.joinList and len(self.joinList):
            return selectTemp + ' FROM ' + self.target + self.build_join() + whereTemp + self.limit + SqlGenerator.build_order(self.orderList) + ";", tuple(params)
        else:
            return selectTemp + ' FROM ' + self.target + whereTemp + self.limit + SqlGenerator.build_order(self.orderList) + ";", tuple(params)

    cdef str build_join(self):
        cdef:
            str joinTemp = "\n"
            JoinCell cell
        for cell in self.joinList:
            foreignKey = cell.key
            joinTemp += "       " + joinTypeMap.get(cell.Type) + \
                        foreignKey.target.tableName + \
                        " on " + foreignKey.owner.tableName + "." + foreignKey.bindCol.name + " = " + \
                        foreignKey.target.tableName + "." + foreignKey.targetBindCol.name + "\n"
        return joinTemp

    @staticmethod
    cdef tuple build_where(FilterListCell whereCol):
        cdef str whereTemp = " WHERE "
        cdef list params = []

        if not whereCol:
            return '', params
        cdef FilterListCell cur = whereCol
        if not cur.next:
            raise SqlInStanceError()
        while True:
            if cur.next and cur.relationship == Relationship.NONE:
                raise SqlMissingRelationshipError(cur.value, str(cur.next))
            whereTemp += SqlGenerator._getWhereCellStr(cur, params)
            if not cur.next:
                break
            cur = cur.next
        return whereTemp, params

    @property
    def joinList(self):
        if not self._joinList:
            self._joinList = []
        return self._joinList

    @staticmethod
    cdef str _getWhereCellStr(FilterListCell cur, list params):
        if cur.col:
            return cur.col.model.tableName + "." + cur.value + " " + relationshipMap.get(cur.relationship, '')
        else:
            params.append(cur.value)
            return " %s " + relationshipMap.get(cur.relationship, '')

cdef class TableGenerator(BaseSqlGenerator):
    def __init__(self, model):
        self.model = model

    cpdef public tuple Build(self):
        cdef:
            str sqlTemp = "CREATE TABLE " + self.model.tableName + "( \n"
            BaseProperty col
            int index

            # 生成列定义语句
        for index in range(len(self.model.col)):
            col = self.model.col[index]
            sqlTemp += TableGenerator.build_col(col)

        # 添加主键定义
        if self.model.pkCol:
            sqlTemp += "   constraint " + self.model.tableName + "_pk PRIMARY KEY (" + self.model.pkCol.name + ")\n"

        # 收尾
        if sqlTemp.endswith(',\n'):
            sqlTemp = sqlTemp[:-2] + '\n'
        sqlTemp += ") ;"
        return sqlTemp,

    @staticmethod
    cdef str build_col(BaseProperty col):
        cdef:
            str temp = "   " + col.name + " "
            str typeArgs = str(col.typeArgs).replace(",", "") if str(col.typeArgs).endswith(",)") else str(col.typeArgs) if len(col.typeArgs) else ''


        # 是否主键
        if col.isPk:
            temp += buildDefault(col)
        else:
            # 数据类型
            temp += (col.targetType + typeArgs + " ")
            # 是否非空
            temp += "NULL " if col.Null else "NOT NULL "
            temp += ("DEFAULT " + buildDefault(col)) if col.hasDefault else ""

        # 额外定义
        temp += col.define + ",\n"

        return temp

cdef str buildDefault(BaseProperty col):
    if isinstance(col._default, BaseSqlGenerator):
        return (<BaseSqlGenerator> col._default).Build()[0]
    elif col.isPk:
        if col._default == AutoIncrement:
            from NonameOrm.DB.DB import DB
            return DB.getInstance().connector.autoFiledAnnounce(col)
        else:
            # 数据类型
            temp = ''
            typeArgs = str(col.typeArgs).replace(",", "") if str(col.typeArgs).endswith(",)") else str(col.typeArgs) if len(col.typeArgs) else ''
            temp += (col.targetType + typeArgs + " ")
            # 是否非空
            temp += "NULL " if col.Null else "NOT NULL "
            return temp

    elif isinstance(col._default, NullDefault) or col._default is NullDefault:
        return ''
    else:
        return _processValue(col.toDBValue(col._default))

cdef _processValue(object data):
    if isinstance(data, bytes):
        return '1' if data == b'\x01' else '0'
    if isinstance(data, str):
        return "'" + data + "'"
    return str(data)

cdef str BuildAlterColSql(str tableName, assignNode: ast.Assign):
    cdef:
        str colName = assignNode.targets[0].id
        str
    return "ALTER TABLE " + tableName + "\n    add " + colName



from NonameOrm.Error.BaseOrmError import BaseOrmError


class SqlInStanceError(BaseOrmError):

    def _process(self, c, t):
        self.msg = f'无法在{c}语句中添加{t}语句块'


class SqlMissingRelationshipError(BaseOrmError):

    def _process(self, str a, str b):
        self.msg = f'在{a}与{b}之间缺少了关系运算符'


class WhereSQLError(BaseOrmError):
    msg = "无法在未指定操作类型的sql上添加where子句"

class SetSQLError(BaseOrmError):
    msg = "无法在非UPDATE语句中添加SET语句"


# cython_ext: language_level=3
# cython: c_string_type=unicode, c_string_encoding=utf8
from NonameOrm.DB.DB cimport DB
from NonameOrm.DB.Generator cimport SqlGenerator
from NonameOrm.Model.DataModel cimport ModelInstance

from NonameOrm.Model.DataModel cimport InstanceList
from .ModelProperty cimport *

cdef class BaseModelExecutor:

    cdef:
        object model
        public SqlGenerator sql
        object work
        DB db
        dict __dict__
        bint executed

    cdef FilterListCell instanceToFilter(self, ModelInstance instance)
    cdef process(self, object res)
    cdef InstanceList processSelect(self, list res)
    cdef processInsert(self, int res)

cdef class ModelExecutor(BaseModelExecutor):
    pass

cdef class AsyncModelExecutor(BaseModelExecutor):
    pass


# cython: c_string_type=unicode, c_string_encoding=utf8
# cython_ext: language_level=3
import json
from typing import Optional

from enum import Enum

from NonameOrm.Error.PropertyError import PropertyVerifyError, PrimaryKeyOverLimitError, PropertyUsageError, \
    ForeignKeyDependError

from NonameOrm.DB.Generator cimport CustomColAnnounce

cdef class NullDefault:
    pass

cdef class AutoIncrement:
    pass

cdef NullDefault _null = NullDefault()

auto_increment = AutoIncrement

current_timestamp = CustomColAnnounce("current_timestamp")
current_date = CustomColAnnounce("current_date")

cdef class FilterListCell:
    def __init__(self, cell = '', Relationship relationship = Relationship.NONE, BaseProperty col = None):
        if <str> cell == '':
            raise
        self.value = cell
        self.relationship = relationship
        self.next = None
        self.col = col

    cpdef FilterListCell append(self, FilterListCell cell, Relationship relationship):
        if self.next:
            return self.next.append(cell, relationship)
        # if self.col and not cell.col:
        #     cell.value = self.col.toDBValue(cell.value)
        self.next = cell
        self.relationship = relationship
        return self

    def check(self):
        if self.next:
            self.next.check()

    def __and__(self, other):
        return self._add(other, AND)

    def __or__(self, other):
        return self._add(other, OR)

    cpdef _add(self, object other, Relationship relationship):
        if issubclass(other.__class__, BaseProperty):
            self.append(FilterListCell(other.name, col=other), relationship)
        elif issubclass(other.__class__, FilterListCell):
            self.append(other, relationship)
        else:
            self.append(FilterListCell(other), relationship)
        return self

cdef class BaseProperty:
    def __init__(self,
                 str name = None,
                 bint pk = False,
                 default = _null,
                 bint Null = True,
                 str define="",
                 tuple typeArgs=(),
                 *args, **kwargs):

        if name:
            self.name = name
        self._Type = self.Type
        self.define = define
        self.isPk = pk
        self.Null = Null
        self._init(*args, **kwargs)
        self._default = default
        self.typeArgs = typeArgs

    def _init(*args, **kwargs):
        pass

    def __set_name__(self, owner, name: str) -> None:
        from NonameOrm import Model
        if not issubclass(owner, Model.DataModel.DataModel):
            raise PropertyUsageError(owner)
        self.name = name
        self.model = owner
        if self.isPk:
            if owner.pkName and owner.pkCol:
                raise PrimaryKeyOverLimitError()
            owner.pkCol = self
            owner.pkName = self.name

    cpdef public bint sizeChecker(self, object value):
        return True

    cpdef public bint verifier(self, object value):
        if isinstance(value, self._Type):
            return self.sizeChecker(value)
        try:
            res = self._Type(value)
            return self.sizeChecker(value)
        except Exception:
            raise PropertyVerifyError(value, self._Type)

    def insertCell(self, value):
        return {
            'col': self,
            'value': value
        }

    def updateCell(self, value):
        return {
            'name': self.name,
            'value': value
        }

    def toDBValue(self, value):
        return value

    def toObjValue(self, value):
        return value

    cdef public str desc(self):
        return f"{self.model.tableName}.{self.name} DESC"

    cdef public str asc(self):
        return f"{self.model.tableName}.{self.name} ASC"

    def __eq__(self, other) -> FilterListCell:
        return self.setFilter(other, EQUAL)

    def __ne__(self, other) -> FilterListCell:
        return self.setFilter(other, NOTEQUAL)

    def __lt__(self, other) -> FilterListCell:
        return self.setFilter(other, SMALLER)

    def __gt__(self, other) -> FilterListCell:
        return self.setFilter(other, BIGGER)

    def __le__(self, other) -> FilterListCell:
        return self.setFilter(other, SMALLER_EQUAL)

    def __ge__(self, other) -> FilterListCell:
        return self.setFilter(other, BIGGER_EQUAL)

    cpdef FilterListCell setFilter(self, object other, Relationship relationship):
        f = FilterListCell(self.name, col=self)
        if issubclass(BaseProperty, other.__class__):
            f.append(FilterListCell(other.name, col=other), relationship)
        else:
            f.append(FilterListCell(self.toDBValue(other)), relationship)
        return f

    @property
    def hasDefault(self):
        return not self._default == _null

    @property
    def Default(self):
        return buildDefault(self)

    @property
    def dbName(self) -> str:
        return self.name

    @property
    def objName(self) -> str:
        return self.name

    @property
    def targetType(self):
        return self._targetType.value

cdef object buildDefault(BaseProperty col):
    if isinstance(col._default, CustomColAnnounce):
        return None
    elif col._default == AutoIncrement:
        return None
    elif isinstance(col._default, NullDefault) or col._default is NullDefault:
        return None
    else:
        return col.toDBValue(col._default)


class intSupportType(Enum):
    Int = 'int'
    bigint = 'int'


class IntProperty(BaseProperty):
    Type: type = int
    supportType: intSupportType = intSupportType

    def _init(self, targetType: intSupportType = intSupportType.Int, *args, **kwargs):
        self._targetType = targetType


class strSupportType(Enum):
    varchar = 'varchar'
    text = 'text'
    longText = 'longtext'
    tinyText = 'tinytext'


class StrProperty(BaseProperty):
    Type: type = str
    supportType: strSupportType = strSupportType

    def _init(self, targetType: strSupportType = strSupportType.varchar, *args, **kwargs):
        self._targetType = targetType
        if not self.typeArgs:
            self.typeArgs = (255,)

    def like(self, str value):
        return self.setFilter(value, LIKE)

    def startsWith(self, str value):
        return self.like(value + "%")

    def endsWith(self, str value):
        return self.like("%" + value)

    def has(self, str value):
        return self.like("%" + value + "%")


class floatSupportType(Enum):
    float = 'float'
    decimal = 'decimal'


class FloatProperty(BaseProperty):
    Type: type = float
    SupportType: floatSupportType = floatSupportType
    size: tuple

    def _init(self, targetType: floatSupportType = floatSupportType.float, tuple size = None, *args, **kwargs):
        self._targetType = targetType
        self.size = size

    def toObjValue(self, value):
        if value is None or value == 'null':
            return None
        return round(float(value), self.typeArgs[1])


class boolSupportType(Enum):
    tinyInt = 'tinyint'
    bit = 'bit'
    varchar = 'varchar'


class BoolProperty(BaseProperty):
    Type: type = bool
    SupportType: boolSupportType = boolSupportType

    def _init(self, targetType: boolSupportType = boolSupportType.tinyInt, *args, **kwargs):
        self._targetType = targetType

    def toDBValue(self, value):
        if self._targetType == boolSupportType.tinyInt:
            return 1 if value else 0
        elif self._targetType == boolSupportType.varchar:
            return 'true' if value else 'false'
        elif self._targetType == boolSupportType.bit:
            return b'\x01' if value else b'\x00'

    def toObjValue(self, object value) -> bool:

        if isinstance(value, bool):
            return value
        if self._targetType == boolSupportType.tinyInt:
            return bool(value)
        elif self._targetType == boolSupportType.varchar:
            if value == 'True' or value == 'true' or value == '1':
                return True
            else:
                return False
        elif self._targetType == boolSupportType.bit:
            return value == b'\x01'


class timestampSupportType(Enum):
    varchar = 'varchar(25)'
    timestamp = 'timestamp'
    datetime = 'datetime'


class TimestampProperty(BaseProperty):
    SupportType: timestampSupportType = timestampSupportType
    Type = str

    def _init(self, targetType: timestampSupportType = timestampSupportType.timestamp, *args, **kwargs):
        self._targetType = targetType

    def toDBValue(self, value):
        return value

    def toObjValue(self, value):
        return value


class jsonSupportType(Enum):
    varchar = 'varchar'
    text = 'text'
    longtext = 'longtext'
    tinytext = 'tinytext'
    json = 'json'


class JsonProperty(BaseProperty):
    Type = dict
    SupportType: jsonSupportType = jsonSupportType

    def _init(self, targetType: jsonSupportType = jsonSupportType.json, *args, **kwargs):
        self._targetType = targetType

    def toObjValue(self, object value):
        if isinstance(value, dict) or isinstance(value, list):
            return value
        return json.loads(value)

    def toDBValue(self, value) -> str:
        if isinstance(value, str) or isinstance(value, bytes):
            return value
        return json.dumps(value)


class ForeignType(Enum):
    ONE_TO_ONE = 1
    ONE_TO_MANY = 2
    MANY_TO_MANY = 3


class ForeignKey(dict):
    __slots__ = ()
    bindCol: BaseProperty
    Type: ForeignType
    targetBindCol: Optional[BaseProperty]
    name: str

    def __init__(self,
                 target=None,
                 Type: ForeignType = ForeignType.ONE_TO_ONE,
                 bindCol: Optional[BaseProperty] = None,
                 targetBindCol: Optional[BaseProperty] = None,
                 middleModel=None
                 ):
        super().__init__()
        self['target'] = target
        self['bindCol'] = bindCol
        self['Type'] = Type
        self['targetBindCol'] = targetBindCol
        self['middleModel'] = middleModel

    def _processMTM(self):
        """
        处理多对多关系
        :return: None
        """
        from NonameOrm.Model.DataModel import MiddleDataModel
        cdef:
            str name = self.owner.__name__.replace('Model', '') + self.target.__name__.replace('Model', '')
            dict attrs = {}

        # 创建主键关联字段
        attrs[self.owner.tableName + '_id'] = self.owner.pkCol.__class__()
        attrs[self.target.tableName + '_id'] = self.target.pkCol.__class__()

        # 创建外键绑定字段
        attrs[self.owner.tableName] = ForeignKey(
            self.owner,
            Type=ForeignType.ONE_TO_MANY,
            bindCol=attrs[self.owner.tableName + '_id'],
            targetBindCol=self.owner.pkCol
        )
        attrs[self.owner.tableName]['name'] = self.owner.tableName
        attrs[self.target.tableName] = ForeignKey(
            self.target,
            Type=ForeignType.ONE_TO_MANY,
            bindCol=attrs[self.target.tableName + '_id'],
            targetBindCol=self.target.pkCol
        )
        attrs[self.target.tableName]['name'] = self.owner.tableName

        # 创建中间模型类
        self['middleModel'] = type(name, (MiddleDataModel,), attrs)

        # 实例化目标模型类外键
        fk = ForeignKey(
            self['middleModel'],
            ForeignType.MANY_TO_MANY,
            bindCol=self.target.pkCol,
            targetBindCol=attrs[self.target.tableName + '_id'],
            middleModel=self['middleModel']
        )
        # 手动给目标表外键绑定所属关系
        fk['owner'] = self.target
        fk['name'] = self.owner.tableName

        # 手动将外键关联添加到对象模型外键列表中
        self.target.fk.append(fk)

        # 手动为对象模型添加外键属性
        setattr(self.target, self.owner.tableName, fk)

        # 更改本外键指向
        self['target'] = self['middleModel']
        self['targetBindCol'] = attrs[self.owner.tableName + '_id']

    def __set_name__(self, owner, name):
        self['name'] = name
        self['owner'] = owner
        from NonameOrm.Model.DataModel import DataModel
        if not issubclass(owner, DataModel):
            raise ForeignKeyDependError()
        if self.Type == ForeignType.MANY_TO_MANY and not self.middleModel:
            self._processMTM()
        elif self.Type == ForeignType.ONE_TO_MANY and self.targetBindCol is None:
            raise RuntimeError("一对多外键需绑定目标表的指定字段")
        if self['target'] is None:
            self['target'] = owner

        if not self.bindCol:
            self['bindCol'] = owner.pkCol
        if not self.targetBindCol:
            self['targetBindCol'] = self.target.pkCol

        self._processTarget()

    def _processTarget(self):
        if self.Type == ForeignType.MANY_TO_MANY:
            pass
        else:
            if hasattr(self.target, self.owner.tableName):
                return
            setattr(self.target, self.owner.tableName, ForeignKey(
                self.owner,
                ForeignType.ONE_TO_ONE,
                bindCol=self.targetBindCol,
                targetBindCol=self.bindCol
            ))

    @property
    def directTarget(self):
        if self.Type == ForeignType.MANY_TO_MANY:
            return self.target.getOtherModelBy(self.owner)
        else:
            return self.target

    def __getattribute__(self, str item):
        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            return object.__getattribute__(self, 'get')(item, None)



from typing import List, Optional, TypedDict, Union

from NonameOrm.Model.DataModel import DataModel, ModelInstance
from NonameOrm.Model.ModelProperty import FilterListCell

from NonameOrm.Ext.Dict import DictPlus


class _Page(TypedDict):
    page: int
    pageSize: int
    content: List[Union[DataModel, ModelInstance]]
    total: int
    totalPage: int


class Page(_Page):
    """
    Page实例类
    本类继承自dict，拥有字典的一切行为

    额外支持通过 . 获取字典内容
    若获取的内容与字典类成员方法,参数相冲突，则优先选取字典内元素
    """

    def __init__(self, page: int, pageSize: int, content=None, total: int = 0):
        pass


class PageAble:
    """
    分页控制器

    本类支持链式调用
    """

    def __init__(self, target: Type[DataModel], page: Optional[int] = 1, pageSize: Optional[int] = 10, findForeign: Optional[bool] = False):
        """

        :param target:
        :param page:
        :param findForeign:
        :param pageSize:
        """
        pass

    def filter(self, args: Optional[FilterListCell] = None) -> "PageAble":
        """
        设置过滤条件

        :param args: FilterListCell
        :return: PageAble
        """
        pass

    def findForeign(self) -> 'PageAble':
        pass

    def orderBy(self, *order: List[str]) -> 'PageAble':
        pass

    def setPage(self, page: int, pageSize: Optional[int] = 0) -> 'PageAble':
        """
        设置分页

        :param page: 当前页数
        :param pageSize: 页面大小 默认可不传
        :return: PageAble
        """
        pass

    def execute(self) -> Page:
        """
        链式调用尽头
        正式进行数据获取

        :return: Page
        """
        pass



# cython_ext: language_level=3
# cython: c_string_type=unicode, c_string_encoding=utf8
from NonameOrm.DB.DB import DB
from NonameOrm.DB.Generator cimport SqlGenerator
from NonameOrm.Model.DataModel cimport InstanceList
from NonameOrm.Model.DataModel import DataModel
from NonameOrm.Model.ModelProperty cimport FilterListCell

from NonameOrm.Ext.Dict cimport DictPlus

cdef class Page(DictPlus):
    """
    Page实例类
    本类继承自dict，拥有字典的一切行为

    额外支持通过 . 获取字典内容
    若获取的内容与字典类成员方法,参数相冲突，则优先选取字典内元素
    """
    def __init__(self, int page, int pageSize, list content=[], int total=0) -> object:
        super().__init__()
        self['page'] = page
        self['pageSize'] = pageSize
        self['content'] = content
        self['total'] = total

cdef class PageAble:
    """
    分页控制器

    本类支持链式调用
    """

    def __init__(self, target, int page=1, int pageSize=10, bint findForeign=False):
        """

        :param target:
        :param page:
        :param findForeign:
        :param pageSize:
        """
        assert pageSize > 0, "pageSize参数必须为大于零的整数"
        assert page > 0, "page参数必须为大于零的整数"

        self.target = target
        self.executor = target.getExecutor()
        self.page = page
        self.pageSize = pageSize
        self.deep = findForeign

    def filter(self, FilterListCell args=None):
        """
        设置过滤条件

        :param args: FilterListCell
        :return: PageAble
        """
        self.executor.sql.select(*self.target.col)
        self.executor.sql.where(args)
        return self

    def findForeign(self):
        self.deep = True
        return self

    def orderBy(self, *order):
        self.executor.sql.orderBy(*order)
        return self

    def setPage(self, int page, int pageSize=0):
        """
        设置分页

        :param page: 当前页数
        :param pageSize: 页面大小 默认可不传
        :return: PageAble
        """
        self.page = page
        if pageSize > 0:
            self.pageSize = pageSize
        return self

    def _getCount(self):
        cdef:
            str sqlTemp
            list args
        sqlTemp, args = SqlGenerator.build_where(self.executor.sql.whereCol)
        return DB.getInstance() \
            .executeSql(f"select count({self.target.pkName}) from {self.target.tableName} {sqlTemp}", args=tuple(args))


    async def _asyncExecute(self):
        cdef:
            Page page = Page(self.page, self.pageSize)  # 创建空page实例


        page.total = (await self._getCount())[0][0]  # 获取当前过滤条件下一共有多少条数据

        # 计算总共有多少页数据
        page.totalPage = int(page.total / self.pageSize) + 1 if page.total % self.pageSize else int(page.total / self.pageSize)

        if not page.total or self.page > page.totalPage:
            # 若分页不可能存在数据则直接返回
            return page

        # 将分页载入Sql中
        self.executor.sql.Limit(self.pageSize, self.pageSize * (self.page - 1))

        page.content = await self.executor.execute()  # 将查询结果放入page实例内

        if self.deep and len(page.content):
            # 判断是否有必要查找外键，若有则进行外键查找
            await self.executor.findListForeignKey(<InstanceList> page.content)

        return page

    cpdef Page _execute(self):
        cdef:
            Page page = Page(self.page, self.pageSize)  # 创建空page实例


        page.total = self._getCount()[0][0]  # 获取当前过滤条件下一共有多少条数据

        # 计算总共有多少页数据
        page.totalPage = int(page.total / self.pageSize) + 1 if page.total % self.pageSize else int(page.total / self.pageSize)

        if not page.total or self.page > page.totalPage:
            # 若分页不可能存在数据则直接返回
            return page

        # 将分页载入Sql中
        self.executor.sql.Limit(self.pageSize, self.pageSize * (self.page - 1))

        page.content = self.executor.execute()  # 将查询结果放入page实例内

        if self.deep and len(page.content):
            # 判断是否有必要查找外键，若有则进行外键查找
            self.executor.findListForeignKey(<InstanceList> page.content)

        return page

    def execute(self):
        """
        链式调用尽头
        正式进行数据获取

        :return: Page
        """
        if DB.getInstance().connector.isAsync:
            return self._asyncExecute()
        else:
            return self._execute()


from asyncio import Task
from typing import Optional, Any, Awaitable

from aiomysql import Connection as AsyncMysqlConnection

from aiosqlite.core import Connection as AsyncSqliteConnection

from NonameOrm.Model.ModelExcutor import AsyncModelExecutor


class BaseConnector:
    """
        数据库连接器基类，无任何功能，需要继承实现
    """
    isAsync: bool

    def getCon(self): ...

    def execute(self, sql: str, con: Any = None, args: tuple = ()): ...

    def process(self, *args, **kwargs):
        pass

    def releaseCon(self, con: Any): ...

    def asyncProcess(self, *args, **kwargs):
        pass


class AioSqliteConnector(BaseConnector):
    """
        异步AioSqlite连接器

        基于 aiosqlite 封装
    """

    async def getCon(self) -> AsyncSqliteConnection: ...

    async def execute(self, sql: str, con: Optional[AsyncSqliteConnection] = None, args: tuple = ()): ...

    def releaseCon(self, con: AsyncMysqlConnection) -> Awaitable: ...

    async def asyncProcess(self,
                           executor:AsyncModelExecutor,
                           ): ...

    pass


class AioMysqlConnector(BaseConnector):
    """
        异步mysql 连接器

        基于 aiomysql 封装
    """
    def __init__(self, *args, **kwargs): ...

    @property
    def config(self) -> dict: ...

    async def getCon(self) -> AsyncMysqlConnection: ...

    def releaseCon(self, con: AsyncMysqlConnection) -> Awaitable: ...

    async def execute(self, sql: str, con: AsyncMysqlConnection = None, args: tuple = ()): ...

    async def asyncProcess(self, *args, **kwargs): ...



# cython: language_level=3
from .BaseOrmError import BaseOrmError


class PropertyUsageError(BaseOrmError):
    def _process(self, type cls, *args, **kwargs) -> None:
        self.msg = f'Property类只能在Model类中被实例化，在{cls.__name__}中则无法正常工作'


class PropertyVerifyError(BaseOrmError):
    def _process(self, data, Type, *args, **kwargs):
        self.msg = f'数据 {data} 并不是 {Type} 类型或无法被 {Type} 类型转化'


class FloatPropertyOutOfRangeError(BaseOrmError):
    def _process(self, size: tuple, value: float, *args, **kwargs):
        self.msg = f'浮点数据整数位限制为{size[0]}位，小数位为{size[1]}位，传入数据{value}超出限制'


class PrimaryKeyOverLimitError(BaseOrmError):
    msg = '每个MODEL对象只允许拥有一个主键约束'

class ForeignKeyDependError(BaseOrmError):
    msg = "外键实例依赖于DataModel子类，无法在非DataModel中创建"


from enum import Enum
from typing import Any, TypedDict, Optional

from NonameOrm.DB.Generator import CustomColAnnounce
from NonameOrm.Model.ModelProperty import FilterListCell

from NonameOrm.Model import DataModel
from NonameOrm.Model.DataModel import MiddleDataModel

current_timestamp = CustomColAnnounce("current_timestamp")
current_date = CustomColAnnounce("current_date")


class InsertCell(TypedDict):
    name: str
    value: Any


class NullDefault: ...


_null: NullDefault


class BaseProperty:
    name: str
    _Type: type
    isPk: bool
    Null: bool
    _default: Any
    targetType: Any
    model: DataModel

    def __init__(self,
                 name: str = None,
                 pk: bool = False,
                 default: Any = _null,
                 Null: bool = True,
                 typeArgs: tuple = (),
                 targetType: Enum = None,
                 *args, **kwargs): ...

    def _init(*args, **kwargs): ...

    def __set_name__(self, owner, name: str) -> None: ...

    def sizeChecker(self, value: Any) -> bool: ...

    def verifier(self, value: Any) -> bool: ...

    def insertCell(self, value) -> InsertCell: ...

    def updateCell(self, value) -> InsertCell: ...

    def toDBValue(self, value) -> Any: ...

    def desc(self) -> str: ...

    def asc(self) -> str: ...

    def toObjValue(self, value): ...

    def __eq__(self, other) -> FilterListCell: ...

    def __ne__(self, other) -> FilterListCell: ...

    def __lt__(self, other) -> FilterListCell: ...

    def __gt__(self, other) -> FilterListCell: ...

    def __le__(self, other) -> FilterListCell: ...

    def __ge__(self, other) -> FilterListCell: ...

    # def setFilter(self, other, relationship) -> FilterListCell: ...

    @property
    def hasDefault(self) -> bool: ...

    @property
    def Default(self) -> Any: ...

    @property
    def dbName(self) -> str: ...

    @property
    def objName(self) -> str: ...


class IntProperty(BaseProperty): ...


class StrProperty(BaseProperty):
    def like(self, value: str) -> FilterListCell: ...

    def startsWith(self, value: str) -> FilterListCell: ...

    def endsWith(self, value: str) -> FilterListCell: ...

    def has(self, value: str) -> FilterListCell: ...


class FloatProperty(BaseProperty): ...


class BoolProperty(BaseProperty): ...


class JsonProperty(BaseProperty): ...


class TimestampProperty(BaseProperty): ...


class timestampSupportType(Enum):
    varchar = 'varchar(25)'
    timestamp = 'timestamp'
    datetime = 'datetime'


class floatSupportType(Enum):
    float = 'float'
    decimal = 'decimal'


class boolSupportType(Enum):
    tinyInt = 'tinyint'
    bit = 'bit'
    varchar = 'varchar(5)'


class jsonSupportType(Enum):
    varchar = 'varchar(255)'
    text = 'text'
    longtext = 'longtext'
    tinytext = 'tinytext'
    json = 'json'


class strSupportType(Enum):
    varchar = 'varchar'
    text = 'text'
    longText = 'longtext'
    tinyText = 'tinytext'


class ForeignType(Enum):
    ONE_TO_ONE = 1
    ONE_TO_MANY = 2
    MANY_TO_MANY = 3


class ForeignKey(dict):
    from NonameOrm import Model
    target: Model.DataModel.DataModel
    bindCol: BaseProperty
    Type: ForeignType
    owner: Model.DataModel.DataModel
    targetBindCol: Optional[BaseProperty]
    middleModel: Optional[MiddleDataModel]
    name: str

    def __init__(self, target, Type: ForeignType = ForeignType.ONE_TO_ONE, bindCol: Optional[BaseProperty] = None, targetBindCol: Optional[BaseProperty] = None): ...

    def __set_name__(self, owner, name): ...

    def __getattribute__(self, item: str): ...

    @property
    def directTarget(self) -> Model.DataModel.DataModel: ...



from typing import Callable, Coroutine

from NonameOrm.DB.Connector import BaseConnector


class DB:
    def __init__(self, connector: BaseConnector, *args, **kwargs): ...

    @classmethod
    def create(cls, *args, **kwargs) -> DB: ...

    @property
    def execute(self) -> Coroutine or Callable: ...

    def GenerateTable(self) -> DB: ...

    @property
    def instance(self) -> DB: ...

    @property
    def connector(self) -> BaseConnector:...

    @property
    def ConnectorType(self): ...

    @classmethod
    def getInstance(cls) -> DB: ...

    @property
    def executeSql(self)-> Callable or Coroutine:...

    @classmethod
    def getInstanceWithCheck(cls) -> DB: ...



# cython_ext: language_level=3
# cython: c_string_type=unicode, c_string_encoding=utf8

from setuptools import setup, Extension, _install_setup_requires
import os

from Cython.Build import cythonize
from os import path as os_path

ex = []

for i, j, k in os.walk('./'):
    for file in k:
        if file.endswith('.pyx'):
            ex.append(Extension('*', [(i if i.endswith('/') else i + '/') + file]))

this_directory = os_path.abspath(os_path.dirname(__file__))


def read_file(filename):
    with open(os_path.join(this_directory, filename), encoding='utf-8') as f:
        long_description = f.read()
    return long_description


def read_requirements(filename):
    return [line.strip() for line in read_file(filename).splitlines()
            if not line.startswith('#')]


setup(
    name='NonameOrm',
    python_requires='>=3.7.0',
    version='1.0.4',  # 包的版本
    author='OToNaShi-4',
    author_email='517431682@qq.com',
    url="https://gitee.com/otonashi-4/python_async_orm",
    description="A Orm support asyncio and sync",
    long_description_content_type="text/markdown",
    packages=["NonameOrm", '.'],
    include_package_data=True,
    keywords=['asyncio', 'aiomysql', 'orm'],
    install_requires=read_requirements('requirements.txt'),
    package_data={
        "NonameOrm": ["*.pyi", "**/*.pyi", "*.py", "*/*.py", "*.pxd", "*/*.pxd", '*/*.pyd', '*.pyd', '*.pyx', '*/*.pyx'],
        '.': ['requirements.txt']
    },
    ext_modules=cythonize(ex, compiler_directives={'language_level': "3"}, ),
    zip_safe=False,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.10',
    ],
)



from typing import List, Tuple

from NonameOrm.Model import DataModel
from NonameOrm.Model.ModelProperty import BaseProperty, FilterListCell, ForeignKey


class JoinCell:
    def __init__(self, fk: ForeignKey, joinType): ...


class BaseSqlGenerator:
    def Build(self): ...


class SqlGenerator(BaseSqlGenerator):
    selectCol: List[BaseProperty]
    limit: str
    joinList: List[JoinCell]
    target: DataModel

    def __init__(self): ...

    def values(self, *args) -> SqlGenerator: ...

    def insert(self, model: DataModel) -> SqlGenerator: ...

    def select(self, *args: List[BaseProperty]) -> SqlGenerator: ...

    def update(self, target: DataModel) -> SqlGenerator: ...

    def set(self, *args) -> SqlGenerator: ...

    def From(self, target: DataModel) -> SqlGenerator: ...

    def where(self, args: FilterListCell = None) -> SqlGenerator: ...

    def Build(self) -> Tuple[str and List[str]]: ...

    def orderBy(self, *args):...

    def Limit(self, count: int, offset: int) -> SqlGenerator: ...

    def join(self, foreignKey: ForeignKey, joinType=None) -> SqlGenerator: ...

    def leftJoin(self, foreignKey: ForeignKey) -> SqlGenerator: ...

    def rightJoin(self, foreignKey: ForeignKey) -> SqlGenerator: ...

    def innerJoin(self, foreignKey: ForeignKey) -> SqlGenerator: ...


class TableGenerator(BaseSqlGenerator):
    model: DataModel

    def Build(self): ...


class CustomColAnnounce(BaseSqlGenerator):
    def __init__(self, sql: str): ...




from typing import Optional, Type, List

from NonameOrm.DB.DB cimport DB
from NonameOrm.Model.ModelExcutor cimport AsyncModelExecutor
from NonameOrm.Model.ModelExcutor import ModelExecutor
from NonameOrm.Model.ModelProperty import ForeignKey
from .ModelProperty cimport *

cdef str get_lower_case_name(str text):
    text = text.replace('Model', '')
    cdef list lst = []
    cdef int index
    cdef str char
    for index, char in enumerate(text):
        if char.isupper() and index != 0:
            lst.append("_")
        lst.append(char)

    return "".join(lst).lower()

cdef class _DataModel:
    @classmethod
    def instanceBuilder(cls: DataModel, *args, **kwargs) -> ModelInstance:
        return cls.modelInstance(*args, **kwargs)


class _DataModelMeta(type):
    ModelList: List['DataModel'] = []

    def __new__(cls, str name, bases: tuple, attrs: dict, **kwargs):
        attrs['col'], attrs['mapping'], attrs['fk'] = _DataModelMeta.getPropertyObj(cls, attrs)
        if not attrs.get('tableName'):
            attrs['tableName'] = get_lower_case_name(name)
        Class = type.__new__(cls, name, bases, attrs)
        _DataModelMeta.ModelList.append(Class)
        setattr(Class, 'modelInstance', _DataModelMeta.buildModelInstance(Class, attrs['col'], name))
        return Class

    @staticmethod
    def buildModelInstance(cls, list cols: List[BaseProperty], str name) -> Type[ModelInstance]:
        cdef dict temp = dict()

        return type(name + 'Instance', (ModelInstance,), {'object': cls, '_temp': {}})

    @staticmethod
    def getPropertyObj(type cls, attrs: dict):
        cdef list temp = []
        cdef list fk = []
        cdef dict mapping = {}
        cdef str key

        for key, item in attrs.items():
            if isinstance(item, BaseProperty):
                temp.append(item)
                mapping[key] = item
            elif isinstance(item, ForeignKey):
                fk.append(item)
                mapping[key] = item
        return temp, mapping, fk


cdef class ModelInstance(dict):
    """
    模型实例基类
    其子类会在DataModel被创建之时自动创建并继承此类
    本类及其子类需要通过实例化模型类时实例化

    本类继承自字典《dict》实例拥有字典的所有能力
    遇到
    """

    def __init__(self, *args, **kwargs):
        cdef:
            str k
            cdef dict data = kwargs



        if len(args):
            if isinstance(args[0], zip):
                super().__init__(args[0])
            elif isinstance(args[0], dict):
                data = <dict> args[0]
                if kwargs.get('check', True):
                    super().__init__()
            else:
                super().__init__()

        from NonameOrm.Model.ModelProperty import ForeignType
        for k, v in self.object.mapping.items():
            if k in data:
                if isinstance(v, ForeignKey) and data[k] is not None:
                    if v.Type == ForeignType.ONE_TO_ONE:
                        self[k] = v.target(data[k], check=kwargs.get('check', True))
                    elif v.Type == ForeignType.ONE_TO_MANY:
                        self[k] = [v.target(i, check=kwargs.get('check', True)) for i in data[k]]
                    elif v.Type == ForeignType.MANY_TO_MANY:
                        self[k] = [v.directTarget(i, check=kwargs.get('check', True)) for i in data[k]]
                elif not kwargs.get('check', True):
                    self[k] = data[k]
                elif isinstance(v, BaseProperty):
                    self[k] = v.toObjValue(data[k])
            elif not kwargs.get('check', True):
                continue
            elif k in self:
                self[k] = v.toObjValue(self[k])
            elif isinstance(v, ForeignKey):
                continue
            else:
                self[v.name] = v.toObjValue(v.Default)

    def __getattribute__(self, str name):
        try:
            return self[name]
        except Exception:
            return object.__getattribute__(self, name)

    def __setattr__(self, key, value):
        if key in self.object.mapping:
            self[key] = value
            return
        raise KeyError()

cdef class InstanceList(list):
    """
    实例列表，内部只有一种对象，就是模型实例的实例
    """

    cdef int len(self):
        return len(self)


class DataModel(_DataModel, metaclass=_DataModelMeta):
    """
    本类为无实例类，无法实例化
    在调用本类的实例化方法时，将会创建本类相对应的模型实例类的实例即

    isinstance(DataModel(), DataModel) -> False
    isinstance(DataModel(), ModelInstance) -> True

    模型实例类 与 模型类 与 模型实例之间的关系需要注意
    """

    _db: Type[DB] = DB
    pkName: Optional[str] = None
    pkCol: Optional[BaseProperty] = None

    def __new__(cls, *args, **kwargs) -> ModelInstance:
        return cls.instanceBuilder(*args, **kwargs)

    @classmethod
    def getAsyncExecutor(cls, work=None) -> AsyncModelExecutor:
        return AsyncModelExecutor(cls, work)

    @classmethod
    def getExecutor(cls, work=None):
        if DB.getInstance().connector.isAsync:
            return cls.getAsyncExecutor(work)
        return ModelExecutor(cls, work)


class MiddleDataModel(DataModel):

    @classmethod
    def instanceBuilder(cls, *args, **kwargs) -> ModelInstance:
        pass

    @classmethod
    def getOtherModelBy(cls, model: DataModel) -> DataModel:
        return cls.getOtherFkBy(model).target

    @classmethod
    def getOtherFkBy(cls, model) -> ForeignKey:
        return cls.fk[0] if cls.fk[0].target is not model else cls.fk[1]



import logging

import faker

from NonameOrm.DB.Connector import AioMysqlConnector, AioSqliteConnector, Sqlite3Connector
from NonameOrm.DB.DB import DB
from NonameOrm.Ext.DataGenerator import data_task, GeneratorRunner
from NonameOrm.Ext.Decorators import use_database
from NonameOrm.Model.DataModel import DataModel
from NonameOrm.Model.ModelProperty import *
import asyncio
from faker import Faker

from NonameOrm.DB.Generator import SqlGenerator

loop = asyncio.new_event_loop()
asyncio.set_event_loop(loop)

from NonameOrm.Ext.PageAble import PageAble

fake = Faker(loacale='zh_CN')

logging.basicConfig(level=logging.DEBUG, format='[ %(levelname)s - %(pathname)s - %(funcName)s] %(message)s')
logger = logging.getLogger(__name__)


class Person(DataModel):
    id = IntProperty(pk=True, Null=False)
    real_name = StrProperty(typeArgs=(8,), Null=False)
    id_card = StrProperty(typeArgs=(24,), Null=False)
    phone = StrProperty(typeArgs=(16,))
    addr = StrProperty(typeArgs=(60,))
    gender = BoolProperty(default=False, Null=False)
    birth_day = TimestampProperty()
    nation = StrProperty(typeArgs=(10,))
    student_id = StrProperty(typeArgs=(18,))
    student_class = StrProperty(typeArgs=(18,))
    specialized_subject = StrProperty(typeArgs=(14,))
    academy = StrProperty(typeArgs=(10,))
    position = StrProperty(typeArgs=(10,))


class User(DataModel):
    id = IntProperty(pk=True, default=auto_increment, Null=False)
    username = StrProperty(typeArgs=(35,), Null=False)
    nickname = StrProperty(typeArgs=(35,), Null=False)
    password = StrProperty(typeArgs=(33,), Null=False)
    avatar = StrProperty(targetType=strSupportType.tinyText)
    email = StrProperty(typeArgs=(100,))
    create_time = StrProperty(typeArgs=(255,))
    update_time = StrProperty(typeArgs=(255,))
    is_delete = BoolProperty(default=False)
    person = ForeignKey(Person)


class abc:

    async def test(self):
        res = await User.getAsyncExecutor().findAllBy()
        print(res)

async def main():
    res = await abc().test()
    print(res)


@data_task(count=10)
def user_generator(time=1):
    name = fake.name()

    return User(
        username=name,
        nickname=fake.name(),
        password=fake.name(),
        email=fake.ascii_email(),
        person=Person(
            phone=fake.phone_number(),
            real_name=fake.name(),
            gender=True,
            id_card=fake.phone_number()
        )
    )


if __name__ == '__main__':
    DB.create(connector=Sqlite3Connector(path=':memory:',showLog=False)).GenerateTable()

    GeneratorRunner().run()

    sql = SqlGenerator().From(User)

    print(User.getExecutor().findAllBy(User.nickname.has('a')))


    # loop.run_until_complete(main())
    pass
